from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..reddit.submission import Submission


class LinkMixin:
    """
    Mixin for items belonging to a submission.
    """

    def __init__(self, link: 'Submission' = None):
        self._link = link

    async def link(self) -> 'Submission':
        """
        Retrieve the submission this item belongs to as a :class:`~apraw.models.Submission`.

        Returns
        -------
        submission: Submission
            The item's parent submission.
        """
        if self._link is None:
            self._link = await self._reddit.submission(getattr(self, "link_id"))
        return self._link

    submission = link



from .enums.distinguishment_option import DistinguishmentOption
from .helpers.apraw_base import aPRAWBase
from .helpers.comment_forest import CommentForest
from .helpers.generator import ListingGenerator
from .helpers.item_moderation import ItemModeration, PostModeration
from .helpers.streamable import Streamable, streamable
from .reddit.comment import Comment, CommentModeration
from .reddit.listing import Listing
from .reddit.message import Message
from .reddit.more_comments import MoreComments
from .reddit.redditor import Redditor
from .reddit.submission import Submission, SubmissionModeration
from .subreddit.banned import BannedUser, BannedListing, SubredditBanned
from .subreddit.moderation import ModAction, SubredditModerator, SubredditModeration
from .subreddit.modmail import ModmailConversation, ModmailMessage, SubredditModmail
from .subreddit.removal_reasons import SubredditRemovalReason, SubredditRemovalReasons
from .subreddit.settings import SubredditSettings
from .subreddit.subreddit import Subreddit
from .subreddit.wiki import WikipageRevision, SubredditWiki, SubredditWikipage
from .user import AuthenticatedUser, Karma, User



"""aPRAW constants."""
from .endpoints import API_PATH, BASE_URL  # noqa: F401

__version__ = "0.6.9"
__tag__ = "alpha"

USER_AGENT_FORMAT = "{} aPRAW v" + __version__ + "-" + __tag__ + "by Dan6erbond"



from typing import TYPE_CHECKING, Any, Dict, Optional

from ..helpers.apraw_base import aPRAWBase
from ...const import API_PATH

if TYPE_CHECKING:
    from .subreddit import Subreddit
    from ..reddit.redditor import Redditor
    from ...reddit import Reddit


class SubredditModmail:
    """
    Helper class to aid in retrieving subreddit modmail.
    """

    def __init__(self, reddit: 'Reddit', subreddit: 'Subreddit'):
        """
        Create an instance of SubredditModmail.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        subreddit: Subreddit
            The subreddit this helper operates under.
        """
        self._reddit = reddit
        self._subreddit = subreddit

    async def __call__(self, id: str, mark_read=False) -> 'ModmailConversation':
        """
        Fetch a :class:`~apraw.models.ModmailConversation` by its ID.

        Parameters
        ----------
        id: str
            The conversation's ID.

        Returns
        -------
        conversation: ModmailConversation
            The conversation requested if it exists.
        """
        return await ModmailConversation(self._reddit, {"id": id}).fetch(mark_read)

    async def conversations(self) -> 'ModmailConversation':
        """
        Retrieve a list of modmail conversations.

        Yields
        ------
        conversation: ModmailConversation
            A modmail conversation held in the subreddit.
        """
        req = await self._reddit.get(API_PATH["modmail_conversations"],
                                     entity=self._subreddit.display_name)
        for _id in req["conversations"]:
            yield ModmailConversation(self._reddit, req["conversations"][_id])


class ModmailConversation(aPRAWBase):
    """
    The model for modmail conversations.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ==================== ==========================================================================================
    Attribute            Description
    ==================== ==========================================================================================
    ``authors``          A list of dictionaries containing authors by name with additional meta information such as
                         ``isMod``, ``isAdmin``, ``isOp``, ``isParticipant``, ``isHidden``, ``id``, ``isDeleted``.
    ``id``               The ID of this conversation.
    ``is_auto``          ``bool``
    ``is_highlighted``   Whether the conversation has been highlighted.
    ``is_internal``      Whether it's an internal mod conversation.
    ``is_repliable``     Whether the conversation can be replied to.
    ``last_mod_update``  A timestamp of the last moderator update or ``None``.
    ``last_unread``      ``None``
    ``last_updated``     A timestamp of the last update made overall.
    ``last_user_update`` A timestamp of the last user update or ``None``.
    ``num_messages``     The number of messages in this conversation.
    ``obj_ids``          A list of dictionaries containing the objects with their IDs and keys.
    ``owner``            A dictionary describing the subreddit this conversation is held in.
    ``participant``      ``Dict``
    ``state``            ``int``
    ``subject``          The subject of this conversation.
    ==================== ==========================================================================================

    """

    def __init__(self, reddit: 'Reddit', data: Dict,
                 owner: 'Subreddit' = None):
        """
        Create an instance of ``ModmailConversation``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        owner: Subreddit
            The subreddit this conversation was held in.
        """
        super().__init__(reddit, data)
        self._owner = owner
        self._messages = list()

    async def fetch(self, mark_read=False):
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        url = API_PATH["modmail_conversation"].format(id=self._data["id"])
        resp = await self._reddit.get(url, mark_read=mark_read)
        self._update(resp)
        return self

    def _update(self, data: Dict[str, Any]):
        """
        Update the base with new information.

        Parameters
        ----------
        data: Dict
            The data obtained from a suitable API endpoint.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        if isinstance(data, Dict):
            if "fields" in data:
                raise Exception(f"{data['message']}: {data['reason']}")
            _data = data.get("conversation", None) or data.get("conversations", None) or data
            super()._update(_data)
            if "messages" in data:
                self._messages = [ModmailMessage(self._reddit, data["messages"][msg_id], self) for msg_id in
                                  data["messages"]]
        else:
            raise Exception(f"Unexpected data: {data}")
        return self

    async def reply(self, body: str, author_hidden: bool = False, internal: bool = False):
        """
        Reply to the modmail conversation.

        Parameters
        ----------
        body: str
            The markdown reply body.
        author_hidden: bool
            Whether the author of this reply should be hidden.
        internal: bool
            Whether the reply is internal.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        url = API_PATH["modmail_conversation"].format(id=self._data["id"])
        resp = await self._reddit.post(url, data={
            "body": body,
            "conversation_id": self._data["id"],
            "isAuthorHidden": author_hidden,
            "isInternal": internal
        })
        return self._update(resp)

    async def _take_action(self, action: str, **kwargs):
        r"""
        Perform an action on the modmail conversation.

        Parameters
        ----------
        action: str
            The action to perform.
        kwargs: \*\*Dict
            Additional request data.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        url = API_PATH["modmail_conversation_action"].format(id=self._data["id"], action=action)
        resp = await self._reddit.post(url, **kwargs)
        return self._update(resp) if resp else self

    async def archive(self):
        """
        Archive the modmail conversation.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        return await self._take_action("archive")

    async def unarchive(self):
        """
        Unarchive the modmail conversation.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        return await self._take_action("unarchive")

    async def highlight(self):
        """
        Highlight the modmail conversation.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        return await self._take_action("highlight")

    async def remove_highlight(self):
        """
        Remove the highlight from the modmail conversation.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        url = API_PATH["modmail_conversation_action"].format(id=self._data["id"], action="higlight")
        resp = await self._reddit.delete(url)
        return self._update(resp) if resp else self

    async def mute(self):
        """
        Mute the modmail conversation.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        return await self._take_action("mute")

    async def unmute(self):
        """
        Unmute the modmail conversation.

        Returns
        -------
        self: ModmailConversation
            The updated model.
        """
        return await self._take_action("unmute")

    async def owner(self) -> 'Subreddit':
        """
        Retrieve the owner subreddit of this conversation.

        Returns
        -------
        owner: Subreddit
            The subreddit this conversation was held in.
        """
        if self._owner is None:
            self._owner = await self._reddit.subreddit(self._data["owner"]["displayName"])
        return self._owner

    async def messages(self) -> 'ModmailMessage':
        """
        Retrieve the messages sent in this conversation.

        Yields
        ------
        message: ModmailMessage
            A message sent in this conversation.
        """
        if not self._messages:
            await self.fetch()

        for msg in self._messages:
            yield msg


class ModmailMessage(aPRAWBase):
    """
    The model for modmail messages.

    Members
    -------
    conversation: ModmailConversation
        The :class:`~apraw.models.ModmailConversation` instance this message belongs to.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ==================== ==================================================
    Attribute            Description
    ==================== ==================================================
    ``id``               The ID of this message.
    ``body``             The HTML body of this message.
    ``body_markdown``    The raw body of this message.
    ``body_md``          An alias to ``body_markdown``.
    ``is_internal``      Whether the message was sent internally.
    ``date``             The datetime string on which the message was sent.
    ==================== ==================================================

    """

    def __init__(self, reddit: 'Reddit', data: Dict, conversation: ModmailConversation):
        """
        Create an instance of a modmail message.

        Parameters
        ----------
        conversation: ModmailConversation
            The :class:`~apraw.models.ModmailConversation` instance this message belongs to.
        data: Dict
            The data obtained from the API.
        """
        super().__init__(reddit, data)
        self.conversation = conversation
        self._author = None

    def _update(self, data: Dict[str, Any]):
        """
        Update the base with new information.

        Parameters
        ----------
        data: Dict
            The data obtained from a suitable API endpoint.

        Returns
        -------
        self: ModmailMessage
            The updated model.
        """
        data["body_md"] = data.get("bodyMarkdown", "")
        super()._update(data)

    async def author(self) -> Optional['Redditor']:
        """
        Retrieve the author of this message as a :class:`~apraw.models.Redditor`.

        Returns
        -------
        author: Redditor or None
            The author of this modmail message if they haven't been deleted yet.
        """
        if self._author is None:
            if not self._data["author"]["isDeleted"]:
                self._author = await self.conversation._reddit.redditor(self._data["author"]["name"])
            else:
                return None
        return self._author



import asyncio
import configparser
import os
from typing import Dict, List, Union, Any

from .endpoints import API_PATH
from .models import (Comment, Listing, Redditor, Submission,
                     Subreddit, User, ListingGenerator, streamable)
from .request_handler import RequestHandler
from .utils import prepend_kind

if os.path.exists('praw.ini'):
    _prawfile = os.path.abspath('praw.ini')
elif 'APPDATA' in os.environ:  # Windows
    _prawfile = os.path.join(os.environ['APPDATA'], 'praw.ini')
elif 'XDG_CONFIG_HOME' in os.environ:  # Modern Linux
    _prawfile = os.path.join(os.environ['XDG_CONFIG_HOME'], 'praw.ini')
elif 'HOME' in os.environ:  # Legacy Linux
    _prawfile = os.path.join(os.environ['HOME'], '.config', 'praw.ini')


class Reddit:
    """
    The Reddit instance with which root requests can be made.

    Members
    -------
    user: User
        An instance of the logged-in Reddit user.
    comment_kind: str
        The prefix that represents :class:`~apraw.models.Comment` in API responses, such as ``t1``.
    account_kind: str
        The prefix that represents :class:`~apraw.models.Redditor` in API responses, such as ``t2``.
    link_kind: str
        The prefix that represents :class:`~apraw.models.Submission` in API responses, such as ``t3``.
    message_kind: str
        The prefix that represents :class:`~apraw.models.Message` in API responses, such as ``t4``.
    subreddit_kind: str
        The prefix that represents :class:`~apraw.models.Subreddit` in API responses, such as ``t5``.
    award_kind: str
        The prefix that represents awards in API responses, such as ``t6``.
    modaction_kind: str
        The prefix that represents :class:`~apraw.models.ModAction` in API responses, such as ``modaction``.
    listing_kind: str
        The prefix that represents :class:`~apraw.models.Listing` in API responses, such as ``listing``.
    wiki_revision_kind: str
        The prefix that represents :class:`~apraw.models.WikipageRevision` in API responses, such as ``WikiRevision``.
    wikipage_kind: str
        The prefix that represents :class:`~apraw.models.SubredditWikipage` in API responses, such as ``wikipage``.
    more_kind: str
        The prefix that represents :class:`~apraw.models.MoreComments` in API responses, such as ``more``.
    request_handler: RequestHandler
        An instance of :class:`~apraw.RequestHandler` with which this Reddit instance will perform HTTP requests.
    """

    def __init__(self, praw_key: str = "", username: str = "", password: str = "",
                 client_id: str = "", client_secret: str = "",
                 user_agent="aPRAW by Dan6erbond"):
        """
        Create a Reddit instance.

        Parameters
        ----------
        praw_key: str
            The key used in a `praw.ini` file instead of manual username, password, client_id and secret.
        username: str
            The Reddit account username.
        password: str
            The Reddit account password.
        client_id: str
            The Reddit script's client_id.
        client_secret
            The Reddit script's client_secret.
        user_agent: str
            User agent to be used in the headers, defaults to "aPRAW by Dan6erbond".
        """
        if praw_key != "":
            config = configparser.ConfigParser()
            config.read(_prawfile)

            self.user = User(self, config[praw_key]["username"], config[praw_key]["password"],
                             config[praw_key]["client_id"], config[praw_key]["client_secret"],
                             config[praw_key]["user_agent"] if "user_agent" in config[praw_key] else user_agent)
        else:
            self.user = User(self, username, password,
                             client_id, client_secret, user_agent)

        self.comment_kind = "t1"
        self.account_kind = "t2"
        self.link_kind = "t3"
        self.message_kind = "t4"
        self.subreddit_kind = "t5"
        self.award_kind = "t6"
        self.modaction_kind = "modaction"
        self.listing_kind = "Listing"
        self.wiki_revision_kind = "WikiRevision"
        self.wikipage_kind = "wikipage"
        self.more_kind = "more"
        self.subreddit_settings_kind = "subreddit_settings"

        self.loop = asyncio.get_event_loop()
        self.request_handler = RequestHandler(self.user)

    #: Streamable listing endpoint.
    @streamable
    def subreddits(self, *args, **kwargs):
        r"""
        A :class:`~apraw.models.ListingGenerator` that returns newly created subreddits, which can be streamed using :code:`reddit.subreddits.stream()`.

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` that retrieves newly created subreddits.
        """
        return ListingGenerator(self, API_PATH["subreddits_new"], *args, **kwargs)

    async def get(self, *args, **kwargs) -> Any:
        """
        Perform an HTTP GET request on the Reddit API.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        kwargs:
            Query parameters to be appended after the URL.

        Returns
        -------
        resp: Any
            The response JSON data.
        """
        return await self.request_handler.get(*args, **kwargs)

    async def post(self, *args, **kwargs) -> Any:
        """
        Perform an HTTP POST request on the Reddit API.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        url: str
            The direct URL to perform the request on.
        data:
            The data to add to the POST body.
        kwargs:
            Query parameters to be appended after the URL.

        Returns
        -------
        resp: Any
            The response JSON data.
        """
        return await self.request_handler.post(*args, **kwargs)

    async def put(self, *args, **kwargs) -> Any:
        """
        Perform an HTTP PUT request on the Reddit API.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        url: str
            The direct URL to perform the request on.
        data:
            The data to add to the POST body.
        kwargs:
            Query parameters to be appended after the URL.

        Returns
        -------
        resp: Any
            The response JSON data.
        """
        return await self.request_handler.put(*args, **kwargs)

    async def delete(self, *args, **kwargs) -> Any:
        """
        Perform an HTTP DELETE request on the Reddit API.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        url: str
            The direct URL to perform the request on.
        kwargs:
            Query parameters to be appended after the URL.

        Returns
        -------
        resp: Any
            The response JSON data.
        """
        return await self.request_handler.post(*args, **kwargs)

    async def get_listing(self, endpoint: str, subreddit: Subreddit = None, kind_filter: List[str] = None,
                          **kwargs) -> Listing:
        r"""
        Retrieve a listing from an endpoint.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        subreddit: Subreddit
            The subreddit to dependency inject into retrieved items when possible.
        kind_filter:
            Kinds to return if given, otherwise all are returned.
        kwargs: \*\*Dict
            Query parameters to be appended after the URL.

        Returns
        -------
        listing: Listing
            The listing containing all the endpoint's children.
        """
        resp = await self.get(endpoint, **kwargs)
        return Listing(self, resp["data"], kind_filter=kind_filter, subreddit=subreddit)

    async def subreddit(self, display_name: str) -> Subreddit:
        """
        Get a `Subreddit` object according to the given name.

        Parameters
        ----------
        display_name: str
            The display name of the subreddit.

        Returns
        -------
        subreddit: Subreddit
            The subreddit if found.
        """
        sub = Subreddit(self, {"display_name": display_name})
        await sub.fetch()
        return sub

    async def info(self, id: str = "", ids: List[str] = [], url: str = ""):
        """
        Get a Reddit item based on its ID or URL.

        Parameters
        ----------
        id: str
            The item's ID.
        ids: List[str]
            Multiple IDs to fetch multiple items at once (max 100).
        url: str
            The item's URL.

        Yields
        -------
        comment: Comment
            A `Comment` object.
        submission: Submission
            A `Submission` object.
        """
        if id:
            for i in await self.get_listing(API_PATH["info"], id=id):
                yield i
        elif ids:
            while ids:
                for i in await self.get_listing(API_PATH["info"], id=",".join(ids[:100])):
                    yield i
                ids = ids[100:]
        elif url:
            for i in await self.get_listing(API_PATH["info"], url=url):
                yield i
        else:
            yield None

    async def submission(self, id: str = "", url: str = "") -> Submission:
        """
        Get a `Submission` object based on its ID or URL.

        Parameters
        ----------
        id: str
            The ID of a submission (with or without kind).
        url: str
            The URL of a submission.

        Returns
        -------
        submission: Submission
            The requested submission.
        """
        if id != "":
            submission = Submission(self, {"id": id})
            await submission.fetch()
            return submission
        else:
            submission = Submission(self, {"url": url})
            await submission.fetch()
            return submission

    async def comment(self, id: str = "", url: str = "") -> Comment:
        """
        Get a `Comment` object based on its ID or URL.

        Parameters
        ----------
        id: str
            The ID of a comment (with or without kind).
        url: str
            The URL of a comment.

        Returns
        -------
        comment: Comment
            The requested comment.
        """
        if id != "":
            comment = Comment(self, {"id": id})
            await comment.fetch()
            return comment
        else:
            comment = Comment(self, {"url": url})
            await comment.fetch()
            return comment

    async def close(self):
        await self.request_handler.close()

    async def redditor(self, username: str) -> Redditor:
        """
        Get a `Redditor` object based the Redditor's username.

        Parameters
        ----------
        username: str
            The Redditor's username (without '/u/').

        Returns
        -------
        redditor: Redditor or None
            The requested Redditor, returns None if not found.
        """
        return await Redditor(self, {"username": username}).fetch()

    async def message(self, to: Union[str, Redditor], subject: str, text: str,
                      from_sr: Union[str, Subreddit] = "") -> bool:
        """
        Message a Redditor or Subreddit.

        Parameters
        ----------
        to: str or Redditor or Subreddit
            The Redditor or Subreddit the message should be sent to.
        subject: str
            The subject of the message.
        text: str
            The text contents of the message.
        from_sr: str or Subreddit
            Optional if the message is being sent from a subreddit.

        Returns
        -------
        result: Dict
            The response JSON data.
        """
        data = {
            "subject": subject,
            "text": text,
            "to": str(to)
        }
        if from_sr != "":
            data["from_sr"] = str(from_sr)
        resp = await self.post(API_PATH["compose"], data=data)
        return not resp["json"]["errors"]



from typing import TYPE_CHECKING, Dict, Any, List, Union

from .comment import Comment
from .submission import Submission
from ..helpers.apraw_base import aPRAWBase
from ..mixins.link import LinkMixin
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit


class MoreComments(aPRAWBase, LinkMixin):
    """
    Represents the model for more comments in a thread.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ============= ================================================================
    Attribute     Description
    ============= ================================================================
    ``count``     The number of comment or more children items in this thread.
    ``name``      The fullname that references this more comments model.
    ``id``        The ID of this more comments model.
    ``parent_id`` The ID of this more comments' parent submission or comment.
    ``depth``     The depth this more comments model goes into.
    ``children``  A list of comment and more comment IDs available in this thread.
    ============= ================================================================

    """

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any], link_id: str):
        """
        Create an instance of ``MoreComments``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        link_id: str
            The ID of the submission this ``MoreComments`` instance belongs to.
            This needs to be known for the ``MoreComments`` instance to be able to fetch its comments.
        """
        data.update(link_id=link_id)
        super().__init__(reddit, data)

        self._comments = []
        self._ids = list(self.children)
        self._index = 0

    async def _next_batch(self):
        """
        Retrieve the next batch of :class:`~apraw.models.Comment` and further ``MoreComments`` in this thread.
        These will be added to the internal list and can be read by using the instance as a generator or fetching the
        comments with :func:`~apraw.models.MoreComments.comments`.
        """
        if not self._ids:
            return

        ids = self._ids[:100]
        self._ids = self._ids[100:]
        resp = await self._reddit.get(API_PATH["morechildren"], **{
            "children": ",".join(ids),
            "link_id": self.link_id,
            "id": self.id,
            "depth": self.depth
        })

        from .listing import MoreChildren
        children = MoreChildren(self._reddit, resp["json"]["data"], [self._reddit.comment_kind, self._reddit.more_kind],
                                self.link_id)
        self._comments.extend(comment for comment in children)

    async def parent(self) -> Union[Submission, Comment]:
        """
        Retrieve the parent submission or comment of this MoreComments object.

        Returns
        -------
        parent: Submission or Comment
            The parent submission or comment of this MoreComments object.
        """
        return await self._reddit.info(self.parent_id)

    async def fetch(self):
        """
        Fetch all the comments in this MoreComments thread.
        """
        while self._ids:
            await self._next_batch()

    def __aiter__(self):
        """
        Permit ``MoreComments`` to operate as an iterator.

        Returns
        -------
        self: MoreComments
            The iterator.
        """
        return self

    async def __anext__(self) -> Comment:
        """
        Permit ``MoreComments`` to operate as a generator.

        Returns the next comment in the listing.

        Returns
        -------
        comment: Comment
            The next comment in the list.

        Raises
        ------
        StopAsyncIteration
            Raised once the list of comments has been exhausted.
        """
        if self._index >= len(self._comments) and not self._ids:
            raise StopAsyncIteration

        if not self._comments or self._ids:
            await self._next_batch()

        self._index += 1
        return self._comments[self._index + 1]

    async def comments(self) -> List[Union[Comment, 'MoreComments']]:
        """
        Retrieve a list of all the :class:`~apraw.models.Comment` and further ``MoreComments`` in this thread.

        Returns
        -------
        comments: List[Union[Comment, MoreComments]]
            A list of all the :class:`~apraw.models.Comment` and further ``MoreComments`` in this thread.
        """
        if not self._comments:
            await self.fetch()

        return self._comments



from typing import TYPE_CHECKING, Dict, List, Union, Any, Optional

from ..helpers.apraw_base import aPRAWBase
from ..helpers.streamable import streamable
from ..reddit.redditor import Redditor
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit
    from .subreddit import Subreddit


class SubredditWiki:
    """
    A helper class to aid in retrieving subreddit wiki pages, revisions as well as creating items.
    """

    def __init__(self, reddit: 'Reddit', subreddit: 'Subreddit'):
        """
        Create an instance of ``SubredditWiki``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        subreddit: Subreddit
            The subreddit this helper performs requests for.
        """
        self._reddit = reddit
        self._subreddit = subreddit
        self._data = None

    #: Streamable listing endpoint.
    @streamable
    def revisions(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to recent wikipage revisions.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.wiki.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to recent wikipage revisions.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["wiki_revisions"].format(sub=self._subreddit), *args,
                                **kwargs)

    async def _fetch(self) -> Dict:
        """
        Fetch the data for a subreddit's wiki from the API.
        """
        url = API_PATH["wiki"].format(sub=self._subreddit)
        self._data = await self._reddit.get(url)
        return self._data

    async def __call__(self) -> List[str]:
        """
        Retrieve a list of the available wikipages.

        Returns
        -------
        pages: List[str]
            A list of all the wikipages in the subreddit by name.
        """
        if not self._data:
            await self._fetch()
        return [page for page in self._data["data"]]

    async def page(self, page: str) -> 'SubredditWikipage':
        """
        Retrieve a specific :class:`~apraw.models.SubredditWikipage` by its name.

        Parameters
        ----------
        page: str
            The wikipage's name which can be retrieved using the list from
            :func:`~apraw.models.SubredditWiki.__call__()`.

        Returns
        -------
        wikipage: SubredditWikipage
            The requested wikipage if it exists.
        """
        url = API_PATH["wiki_page"].format(sub=self._subreddit, page=page)
        resp = await self._reddit.get(url)
        return SubredditWikipage(page, self._reddit, self._subreddit, resp["data"])

    async def create(self, page: str, content_md: str = "", reason: str = "") -> 'SubredditWikipage':
        """
        Create a new wikipage on the subreddit.

        Parameters
        ----------
        page: str
            The wikipage's name.
        content_md: str
            The wikipage's content as a markdown string.
        reason: str
            An optional string detailing the reason for the creation of this wikipage.

        Returns
        -------
        wikipage: SubredditWikipage
            The newly created wikipage.
        """
        resp = await self._reddit.post(
            API_PATH["wiki_edit"].format(sub=self._subreddit), data={
                "page": page,
                "content": content_md,
                "reason": reason
            })
        return resp if resp else await self.page(page)


class SubredditWikipage(aPRAWBase):
    """
    The model that represents Subreddit wikipages.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================= =============================================================
    Attribute         Description
    ================= =============================================================
    ``content_html``  The content's of the wikipage as an HTML string.
    ``content_md``    The content's of the wikipage formatted as a markdown string.
    ``may_revise``    ``bool``
    ``name``          The wikipage's name.
    ``reason``        The reason text for the wikipage's current revision.
    ``revision_by``   The author of the wikipage's current revision.
    ``revision_date`` The date on which the current revision was made.
    ``revision_id``   The ID of the wikipage's current revision.
    ================= =============================================================
    """

    def __init__(self, name: str, reddit: 'Reddit', subreddit: 'Subreddit', data: Dict = None):
        """
        Create an instance of ``SubredditWikipage``.

        Parameters
        ----------
        name: str
            The wikipage's name.
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        subreddit: Subreddit
            The subreddit this helper performs requests for.
        data: Dict
            The data returned by the API endpoint.
        """
        super().__init__(reddit, data, reddit.wikipage_kind)

        self.name = name
        self._subreddit = subreddit

    #: Streamable listing endpoint.
    @streamable
    def revisions(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to fetch specific wikipage revisions.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.wiki.page("test").stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to fetch specific wikipage revisions.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["wiki_page_revisions"].format(
            sub=self._subreddit, page=self.name), *args, **kwargs)

    async def _alloweditor(self, username: str, act: str):
        """
        Remove or add a Redditor from the editors of this wikipage.

        Parameters
        ----------
        username: str
            The Redditor's username without the prefix.
        act: str
            Either "add" or "del" based on the act to be performed.

        Returns
        -------
        res: bool or Any
            ``True`` if the request was successful, otherwise the response's raw data.
        """
        resp = await self._reddit.post(
            API_PATH["wiki_alloweditor"].format(sub=self._subreddit, act=act), data={
                "page": self.name,
                "username": username
            })
        return True if not resp else resp

    async def add_editor(self, username: str) -> Union[bool, Any]:
        """
        Add a Redditor to the editors of this wikipage.

        Parameters
        ----------
        username: str
            The Redditor's username without the prefix.

        Returns
        -------
        res: bool or Any
            ``True`` if the request was successful, otherwise the response's raw data.
        """
        return await self._alloweditor(username, "add")

    async def del_editor(self, username: str) -> Union[bool, Any]:
        """
        Remove a Redditor from the editors of this wikipage.

        Parameters
        ----------
        username: str
            The Redditor's username without the prefix.

        Returns
        -------
        res: bool or Any
            ``True`` if the request was successful, otherwise the response's raw data.
        """
        return await self._alloweditor(username, "del")

    async def edit(self, content_md: str, reason: Optional[str] = "") -> Union[bool, Any]:
        """
        Edit a wikipage's markdown contents.

        Parameters
        ----------
        content_md: str
            The new wikipage's content as a markdown string.
        reason: Optional[str]
            An optional reason for this edit.

        Returns
        -------
        res: bool or Any
            ``True`` if the request was successful, otherwise the response's raw data.
        """
        resp = await self._reddit.post(
            API_PATH["wiki_edit"].format(sub=self._subreddit), data={
                "page": self.name,
                "content": content_md,
                "reason": reason
            })
        return resp if resp else True

    async def hide(self, revision: Union[str, 'WikipageRevision']):
        """
        Hide a wikipage revision from the public history.

        Parameters
        ----------
        revision: str or WikipageRevision
            The wikipage revision either as a :class:`~apraw.models.WikipageRevision` or its ID string.

        Returns
        -------
        res: bool or Any
            ``True`` if the request was successful, otherwise the response's raw data.
        """
        resp = await self._reddit.post(
            API_PATH["wiki_hide"].format(sub=self._subreddit), data={
                "page": self.name,
                "revision": str(revision)
            })
        return resp if resp else True

    async def revert(self, revision: Union[str, 'WikipageRevision']) -> Union[bool, Any]:
        """
        Revert a wikipage to its previous revision.

        Parameters
        ----------
        revision: str or WikipageRevision
            The wikipage revision either as a :class:`~apraw.models.WikipageRevision` or its ID string.

        Returns
        -------
        res: bool or Any
            ``True`` if the request was successful, otherwise the response's raw data.
        """
        resp = await self._reddit.post(
            API_PATH["wiki_revert"].format(sub=self._subreddit), data={
                "page": self.name,
                "revision": str(revision)
            })
        return resp if resp else True


class WikipageRevision(aPRAWBase):
    """
    The model that represents wikipage revisions.

    Members
    -------
    author: Redditor
        The Redditor that made this revision.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    =================== ====================================================
    Attribute           Description
    =================== ====================================================
    ``timestamp``       A timestamp of when the revision was made.
    ``page``            The name of the page the revision addresses.
    ``revision_hidden`` Whether the revision has been hidden by the editors.
    ``reason``          The reason string for this revision if available.
    ``id``              The ID of this revision.
    =================== ====================================================
    """

    def __init__(self, reddit: 'Reddit', data: Dict = None):
        """
        Create an instance of ``WikipageRevision``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data returned by the API endpoint.
        """
        super().__init__(reddit, data, reddit.wiki_revision_kind)

        self.author = Redditor(reddit, data["author"]["data"])

    def __str__(self):
        """
        Retrieve the string representation of this object.

        Returns
        -------
        id: str
            The ID of this wikipage revision.
        """
        return self.id



from enum import Enum


class DistinguishmentOption(Enum):
    """
    An enum for the distinguishment types.

     - YES | "yes"
     - NO | "no"
     - ADMIN | "admin"
     - SPECIAL | "special"
    """
    YES = "yes"
    NO = "no"
    ADMIN = "admin"
    SPECIAL = "special"



from typing import TYPE_CHECKING, Dict, Optional, Any, Union

from ..helpers.apraw_base import aPRAWBase
from ...const import API_PATH

if TYPE_CHECKING:
    from .subreddit import Subreddit
    from ...reddit import Reddit


class SubredditRemovalReason(aPRAWBase):
    """
    The model representing subreddits.

    Members
    -------
    url: str
        The API URL to this specific removal reason.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    =========== ============================================================
    Attribute   Description
    =========== ============================================================
    ``message`` The message for this removal reason that is sent to authors.
    ``id``      The ID of this removal reason.
    ``title``   The title of this removal reason in the subreddit.
    =========== ============================================================
    """

    def __init__(self, reddit: 'Reddit', subreddit: 'Subreddit', data: Dict):
        """
        Create an instance of a ``SubredditRemovalReason``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        subreddit: Subreddit
            The subreddit this helper operates under.
        data: Dict
            The data obtained from the /about endpoint.
        """
        super().__init__(reddit, data)
        self._subreddit = subreddit
        self.url = API_PATH["subreddit_removal_reason"].format(sub=self._subreddit.display_name, id=self.id)

    async def fetch(self):
        """
        Fetch the data for this removal reason. The :class:`~apraw.models.aPRAWBase` class will automatically update
        and/or add members returned by the API.
        """
        url = API_PATH["subreddit_removal_reasons"].format(sub=self._subreddit.display_name)
        res = await self._reddit.get(url)
        super()._update(res["data"][self.id])

    async def delete(self) -> Any:
        """
        Delete this removal reason from the subreddit.

        Returns
        -------
        response: Any
            The API endpoint raw response.
        """
        res = await self._reddit.delete(self.url)
        return res

    async def update(self, title: Optional[str] = None, message: Optional[str] = None) -> Any:
        """
        Update the title and/or message of this removal reason.

        Parameters
        ----------
        title: Optional[str]
            The updated title for this removal reason. If none is specified the original title will be reused.
        message: Optional[str]
            The updated message for this removal reason. If none is specified the original message will be reused.

        Returns
        -------
        response: Any
            The API endpoint raw response.
        """
        data = {
            k: v if v else getattr(self, k)
            for k, v in {"message": message, "title": title}.items()
        }
        await self._reddit.put(self.url, data=data)

    def __str__(self):
        """
        Retrieve a string representation of this removal reason.

        Returns
        -------
        id: str
            This removal reason's full ID.
        """
        return self.id


class SubredditRemovalReasons:
    """
    A helper to aid in retrieving and adding removal reasons to a subreddit.
    """

    def __init__(self, reddit: 'Reddit', subreddit: 'Subreddit'):
        """
        Create an instance of ``SubredditRemovalReasons``.

        Parameters
        ----------
        reddit: Reddit
            The reddit instance with which requests are made.
        subreddit: Subreddit
            The subreddit for which this helper performs requests.
        """
        self._reddit = reddit
        self._subreddit = subreddit
        self._removal_reasons = list()
        self._index = 0

    async def _fetch(self):
        """
        Refresh the data for the subreddit removal reasons.
        """
        url = API_PATH["subreddit_removal_reasons"].format(sub=self._subreddit.display_name)
        res = await self._reddit.get(url)

        for reason_id in res["order"]:
            reason = SubredditRemovalReason(self._reddit, self._subreddit, res["data"][reason_id])
            self._removal_reasons.append(reason)

    async def get(self, item: Union[int, str]) -> SubredditRemovalReason:
        """
        Retrieve a removal reason based on its ID or index.

        Parameters
        ----------
        item: int or str
            The item's ID or index.

        Returns
        -------
        reason: SubredditRemovalReason
            The removal reason that was found in the list.

        Raises
        ------
        StopIteration
            If no removal reason by the given ID was found.
        IndexError
            If the index given doesn't exist in the list of removal reasons.
        """
        if not self._removal_reasons:
            await self._fetch()

        if isinstance(item, str):
            return next(reason for reason in self._removal_reasons if reason.id == item)
        else:
            return self._removal_reasons[item]

    def __aiter__(self):
        """
        Permit ``SubredditRemovalReasons`` to operate as an iterator.

        Returns
        -------
        self: SubredditRemovalReasons
            The iterator.
        """
        return self

    async def __anext__(self):
        """
        Permit ``SubredditRemovalReason`` to operate as a generator.

        Returns the next removal reason in the listing.

        Returns
        -------
        reason: SubredditRemovalReason
            The next removal reason in the list.

        Raises
        ------
        StopAsyncIteration
            Raised once the list of removal reasons has been exhausted.
        """
        if not self._removal_reasons:
            await self._fetch()

        if self._index >= len(self._removal_reasons):
            raise StopAsyncIteration

        self._index += 1
        return self._removal_reasons[self._index - 1]

    async def add(self, title: str, message: str) -> SubredditRemovalReason:
        """
        Add a removal reason to the subreddit's list.

        Parameters
        ----------
        title: str
            The title under which this removal reason is saved.
        message: str
            The message that is sent to author's when the removal reason is used.

        Returns
        -------
        reason: SubredditRemovalReason
            The newly created, and fetched, removal reason.
        """
        data = {"message": message, "title": title}
        url = API_PATH["subreddit_removal_reasons"].format(sub=self._subreddit.display_name)

        data = await self._reddit.post(url, data=data)

        reason = SubredditRemovalReason(self._reddit, self._subreddit, data)
        await reason.fetch()
        return reason
